from datetime import date, datetime, timezone

import prefab_pb2 as Prefab


class ConfigValueWrapper:
    @staticmethod
    def wrap(value, confidential=None):
        value_type = type(value)
        if value_type == Prefab.ConfigValue:
            return value
        elif value_type == int:
            return Prefab.ConfigValue(int=value, confidential=confidential)
        elif value_type == float:
            return Prefab.ConfigValue(double=value, confidential=confidential)
        elif value_type == bool:
            return Prefab.ConfigValue(bool=value, confidential=confidential)
        elif value_type == list:
            return Prefab.ConfigValue(
                string_list=Prefab.StringList(values=[str(x) for x in value]),
                confidential=confidential,
            )
        elif value_type == datetime:
            return Prefab.ConfigValue(
                string=ConfigValueWrapper._format_date_time(value),
                confidential=confidential,
            )
        elif value_type == date:
            return Prefab.ConfigValue(
                string=ConfigValueWrapper._format_date_time(
                    datetime.combine(value, datetime.min.time(), timezone.utc)
                ),
                confidential=confidential,
            )
        else:
            return Prefab.ConfigValue(string=value, confidential=confidential)

    @staticmethod
    def _format_date_time(value: datetime):
        return value.strftime("%Y-%m-%dT%H:%M:%SZ")



import threading
from typing import Optional


class CountDownLatch:
    def __init__(self, count: int = 1) -> None:
        self.count = count
        self.lock = threading.Condition()

    def count_down(self) -> None:
        with self.lock:
            self.count -= 1
            if self.count <= 0:
                self.lock.notify_all()

    def wait(self, timeout: Optional[int] = None) -> bool:
        with self.lock:
            if self.count > 0:
                self.lock.wait(timeout)
            return self.count <= 0



from __future__ import annotations
import functools
import time
from collections.abc import Sequence

from .read_write_lock import ReadWriteLock
from .config_value_unwrapper import ConfigValueUnwrapper
from .context import Context
from ._internal_logging import InternalLogger
from .simple_criterion_evaluators import (
    NumericOperators,
    StringOperators,
    DateOperators,
    SemverOperators,
    RegexMatchOperators,
)
import prefab_pb2 as Prefab

logger = InternalLogger(__name__)


class ConfigResolver:
    def __init__(self, base_client, config_loader):
        self.local_store = None
        self.lock = ReadWriteLock()
        self.base_client = base_client
        self.config_loader = config_loader
        self.project_env_id = 0
        self.default_context = {}
        self.make_local()

    def get(self, key, context=None) -> "Evaluation | None":
        with self.lock.read_locked():
            raw_config = self.raw(key)

        if raw_config is None:
            merged_context = self.evaluation_context(context)
            return Evaluation(
                config=None,
                value=None,
                config_row_index=0,
                value_index=0,
                context=merged_context,
                resolver=self,
            )
        else:
            return self.evaluate(raw_config, context=context)

    def raw(self, key) -> Prefab.ConfigValue | None:
        via_key = self.local_store.get(key)
        if via_key is not None:
            return via_key["config"]
        return None

    def evaluate(self, config, context=None) -> "Evaluation | None":
        return CriteriaEvaluator(
            config,
            project_env_id=self.project_env_id,
            resolver=self,
            base_client=self.base_client,
        ).evaluate(self.evaluation_context(context))

    def evaluation_context(self, context):
        merged_context = Context()
        merged_context.merge_context_dict(self.base_client.global_context.to_dict())
        merged_context.merge_context_dict(self.default_context)
        if Context.get_current():
            merged_context.merge_context_dict(Context.get_current().to_dict())
        if context:
            merged_context.merge_context_dict(
                Context.normalize_context_arg(context).to_dict()
            )
        return merged_context

    def update(self):
        self.make_local()

    def make_local(self):
        with self.lock.write_locked():
            self.local_store = self.config_loader.calc_config()

    @property
    def default_context(self):
        return self._default_context

    @default_context.setter
    def default_context(self, value):
        self._default_context = Context.normalize_context_arg(value).to_dict()


OPS = Prefab.Criterion.CriterionOperator


class CriteriaEvaluator:
    def __init__(self, config, project_env_id, resolver, base_client):
        self.config = config
        self.project_env_id = project_env_id
        self.resolver = resolver
        self.base_client = base_client

    def evaluate(self, props):
        matching_env_row_values = self.matching_environment_row_values()
        default_row_index = 1 if matching_env_row_values else 0
        for value_index, conditional_value in enumerate(matching_env_row_values):
            if self.all_criteria_match(conditional_value, props):
                return Evaluation(
                    self.config,
                    conditional_value.value,
                    value_index,
                    0,
                    props,
                    self.resolver,
                )
        for value_index, conditional_value in enumerate(self.default_row_values()):
            if self.all_criteria_match(conditional_value, props):
                return Evaluation(
                    self.config,
                    conditional_value.value,
                    value_index,
                    default_row_index,
                    props,
                    self.resolver,
                )
        return None

    def all_criteria_match(self, conditional_value, props):
        for criterion in conditional_value.criteria:
            if not self.evaluate_criterion(criterion, props):
                return False
        return True

    def evaluate_criterion(self, criterion, properties):
        if criterion.property_name in ["prefab.current-time", "reforge.current-time"]:
            value_from_properties = int(time.time() * 1000)
        else:
            value_from_properties = properties.get(criterion.property_name)

        deepest_value = ConfigValueUnwrapper.deepest_value(
            criterion.value_to_match, self.config, properties
        )

        if criterion.operator in [OPS.LOOKUP_KEY_IN, OPS.PROP_IS_ONE_OF]:
            return self.one_of(criterion, value_from_properties, properties)
        if criterion.operator in [OPS.LOOKUP_KEY_NOT_IN, OPS.PROP_IS_NOT_ONE_OF]:
            return not self.one_of(criterion, value_from_properties, properties)
        if criterion.operator == OPS.IN_SEG:
            return self.in_segment(criterion, properties)
        if criterion.operator == OPS.NOT_IN_SEG:
            return not self.in_segment(criterion, properties)
        if criterion.operator in StringOperators.SUPPORTED_OPERATORS:
            return StringOperators.evaluate(
                value_from_properties, criterion.operator, deepest_value.unwrap()
            )
        if criterion.operator == OPS.HIERARCHICAL_MATCH:
            return value_from_properties.startswith(criterion.value_to_match.string)
        if criterion.operator == OPS.ALWAYS_TRUE:
            return True
        if criterion.operator in DateOperators.SUPPORTED_OPERATORS:
            return DateOperators.evaluate(
                value_from_properties, criterion.operator, deepest_value.unwrap()
            )
        if criterion.operator in NumericOperators.SUPPORTED_OPERATORS:
            return NumericOperators.evaluate(
                value_from_properties, criterion.operator, deepest_value.unwrap()
            )
        if criterion.operator in RegexMatchOperators.SUPPORTED_OPERATORS:
            return RegexMatchOperators.evaluate(
                value_from_properties, criterion.operator, deepest_value.unwrap()
            )
        if criterion.operator in SemverOperators.SUPPORTED_OPERATORS:
            return SemverOperators.evaluate(
                value_from_properties, criterion.operator, deepest_value.unwrap()
            )

        logger.info(f"Unknown criterion operator {criterion.operator}")
        return False

    @staticmethod
    def negate(negate, value):
        return not value if negate else value

    @staticmethod
    def _ensure_list(value):
        if isinstance(value, list):
            return value
        if isinstance(value, Sequence) and not isinstance(value, (str, bytes)):
            return list(value)  # Convert other sequences to lists
        return [value]  # Wrap everything else in a list

    def one_of(self, criterion, value, properties):
        criterion_value_or_values = ConfigValueUnwrapper.deepest_value(
            criterion.value_to_match, self.config, properties
        ).unwrap()

        criterion_values = self._ensure_list(criterion_value_or_values)
        values = self._ensure_list(value)

        return any(str(v1) == str(v2) for v1 in criterion_values for v2 in values)

    def in_segment(self, criterion, properties):
        return (
            self.resolver.get(criterion.value_to_match.string, context=properties)
            .raw_config_value()
            .bool
        )

    def matching_environment_row_values(self):
        env_rows = [
            row for row in self.config.rows if row.project_env_id == self.project_env_id
        ]
        if env_rows == []:
            return []
        else:
            return env_rows[0].values

    def default_row_values(self):
        env_rows = [
            row for row in self.config.rows if row.project_env_id != self.project_env_id
        ]
        if env_rows == []:
            return []
        else:
            return env_rows[0].values


class Evaluation:
    def __init__(
        self,
        config: Prefab.Config | None,
        value: Prefab.ConfigValue | None,
        value_index: int,
        config_row_index: int,
        context: Context,
        resolver: ConfigResolver,
    ):
        self.config = config
        self.value = value
        self.value_index = value_index
        self.config_row_index = config_row_index
        self.context = context
        self.resolver = resolver

    def unwrapped_value(self):
        return self.deepest_value().unwrap()

    def raw_config_value(self):
        return self.value

    @functools.cache
    def deepest_value(self):
        return ConfigValueUnwrapper.deepest_value(
            self.value, self.config, self.resolver, self.context
        )



import threading
from contextlib import contextmanager


# source: https://www.oreilly.com/library/view/python-cookbook/0596001673/ch06s04.html
class ReadWriteLock:
    """A lock object that allows many simultaneous "read locks", but
    only one "write lock." """

    def __init__(self):
        self._read_ready = threading.Condition(threading.Lock())
        self._readers = 0

    def try_acquire_read(self, timeout):
        acquired = self._read_ready.acquire(timeout=timeout)
        if not acquired:
            print("not acquired, returning False")
            return False
        self._acquire_read_core()
        return True

    def acquire_read(self):
        """Acquire a read lock. Blocks only if a thread has
        acquired the write lock."""
        if self._read_ready.acquire():
            self._acquire_read_core()

    def _acquire_read_core(self):
        try:
            self._readers += 1
        finally:
            self._read_ready.release()

    def release_read(self):
        """Release a read lock."""
        self._read_ready.acquire()
        try:
            self._readers -= 1
            if not self._readers:
                self._read_ready.notify_all()
        finally:
            self._read_ready.release()

    @contextmanager
    def read_locked(self):
        """This method is designed to be used via the `with` statement."""
        try:
            self.acquire_read()
            yield
        finally:
            self.release_read()

    @contextmanager
    def read_locked_timeout(self, timeout=1):
        """This method is designed to be used via the `with` statement.
        check return value: returns True only if lock is acquired
        """
        result = self.try_acquire_read(timeout=timeout)
        try:
            if result:
                yield result
            else:
                # return the result even if false rather than throw an exception, caller needs to check the result
                yield result

        finally:
            if result:
                self.release_read()

    def acquire_write(self):
        """Acquire a write lock. Blocks until there are no
        acquired read or write locks."""
        self._read_ready.acquire()
        while self._readers > 0:
            self._read_ready.wait()

    def release_write(self):
        """Release a write lock."""
        self._read_ready.release()

    @contextmanager
    def write_locked(self):
        """This method is designed to be used via the with statement"""
        try:
            self.acquire_write()
            yield
        finally:
            self.release_write()



import prefab_pb2 as Prefab

from sdk_reforge._internal_constants import LOG_LEVEL_BASE_KEY


class MissingFeatureFlagValueException(Exception):
    "Raised when a feature flag is missing a value"

    def __init__(self, key, source):
        super().__init__(
            "Feature flag config `%s` in %s must have a `value`" % (key, source)
        )


class ConfigParser:
    def parse(key, value, config, source):
        if isinstance(value, dict):
            return ConfigParser.parse_dict(key, value, config, source)
        else:
            parsed = ConfigParser.parse_scalar(key, value, config, source)
            config[key] = parsed
            return config

    def parse_dict(key, value, config, source):
        if value.get("feature_flag") is not None:
            config[key] = ConfigParser.feature_flag_config(key, value, source)
        elif value.get("type") == "provided":
            config[key] = ConfigParser.provided_config(key, value, source)
        else:
            for nest_key in value:
                nest_value = value[nest_key]
                if nest_key == "_":
                    nested_key = key
                else:
                    nested_key = "%s.%s" % (key, nest_key)
                config = ConfigParser.parse(nested_key, nest_value, config, source)
        return config

    def parse_scalar(key, value, config, source):
        return {
            "source": source,
            "match": "default",
            "config": Prefab.Config(
                config_type="CONFIG",
                key=key,
                rows=[
                    Prefab.ConfigRow(
                        values=[
                            Prefab.ConditionalValue(
                                value=ConfigParser.value_from(key, value)
                            )
                        ]
                    )
                ],
            ),
        }

    def provided_config(key, value, source):
        value = Prefab.ConfigValue(
            provided=Prefab.Provided(source="ENV_VAR", lookup=value["lookup"]),
            confidential=value.get("confidential"),
        )

        row = Prefab.ConfigRow(values=[Prefab.ConditionalValue(value=value)])

        return {
            "source": source,
            "match": value.provided.lookup,
            "config": Prefab.Config(config_type="CONFIG", key=key, rows=[row]),
        }

    def feature_flag_config(key, value, source):
        if "value" not in value.keys():
            raise MissingFeatureFlagValueException(key, source)

        if value.get("criterion"):
            criteria = [ConfigParser.parse_criterion(value["criterion"])]
        else:
            criteria = []

        variant = ConfigParser.feature_flag_variant_from(key, value["value"])

        row = Prefab.ConfigRow(
            values=[
                Prefab.ConditionalValue(
                    criteria=criteria,
                    value=Prefab.ConfigValue(
                        weighted_values=Prefab.WeightedValues(
                            weighted_values=[
                                Prefab.WeightedValue(weight=1000, value=variant)
                            ]
                        )
                    ),
                )
            ]
        )

        return {
            "source": source,
            "match": key,
            "config": Prefab.Config(
                config_type="FEATURE_FLAG",
                key=key,
                rows=[row],
                allowable_values=[variant],
            ),
        }

    def parse_criterion(criterion):
        return Prefab.Criterion(
            operator=criterion.get("operator"),
            property_name=ConfigParser.parse_property(criterion),
            value_to_match=ConfigParser.parse_value_to_match(criterion.get("values")),
        )

    def parse_property(criterion):
        if criterion.get("operator") == "LOOKUP_KEY_IN":
            return "LOOKUP"
        else:
            return criterion.get("property")

    def parse_value_to_match(values):
        if isinstance(values, list):
            return Prefab.ConfigValue(string_list=Prefab.StringList(values=values))
        else:
            raise "can't handle %s" % values

    def value_from(key, value):
        if isinstance(value, str):
            if key.startswith(LOG_LEVEL_BASE_KEY):
                log_level = ConfigParser.parse_log_level(value)
                return {"log_level": log_level}
            else:
                return {"string": value}
        elif isinstance(value, bool):
            return {"bool": value}
        elif isinstance(value, int):
            return {"int": value}
        elif isinstance(value, float):
            return {"double": value}
        else:
            pass

    def feature_flag_variant_from(key, value):
        if isinstance(value, str):
            if key.startswith(LOG_LEVEL_BASE_KEY):
                log_level = ConfigParser.parse_log_level(value)
                return Prefab.ConfigValue(log_level=log_level)
            else:
                return Prefab.ConfigValue(string=value)
        elif isinstance(value, bool):
            return Prefab.ConfigValue(bool=value)
        elif isinstance(value, int):
            return Prefab.ConfigValue(int=value)
        elif isinstance(value, float):
            return Prefab.ConfigValue(double=value)
        else:
            pass

    def parse_log_level(log_level):
        if log_level.upper() in Prefab.LogLevel.keys():
            return log_level.upper()
        else:
            return "NOT_SET_LOG_LEVEL"



"""
Reforge Python client library.

This module provides access to the Reforge configuration and feature flag service.

Main components:
- ReforgeSDK: The main SDK for interacting with Reforge
- Options: Configuration options for the SDK
- Context: Context information for evaluating configs and feature flags

Re-exported Protocol Buffer types:
- ConfigValue: Represents a configuration value
- StringList: Represents a list of strings
- ProtoContext: Protocol buffer Context class
- ContextSet: Collection of contexts
- ContextShape: Shape information for contexts
- LogLevel: Enumeration of log levels
- Json: Represents JSON data in configuration values
- Schema: Represents schema validation for configuration values
"""

from typing import Optional

from . import _internal_logging
from .options import Options as Options
from .sdk import ReforgeSDK as ReforgeSDK
from importlib.metadata import version
from .read_write_lock import ReadWriteLock as _ReadWriteLock
from .context import Context, NamedContext
from .feature_flag_sdk import FeatureFlagSDK
from .config_sdk import ConfigSDK
from .constants import (
    ConfigValueType,
    ContextDictType,
    ContextDictOrContext,
    NoDefaultProvided,
)

# Re-export Protocol Buffer types for easier access
import prefab_pb2
from prefab_pb2 import (
    ConfigValue,
    StringList,
    Context as ProtoContext,
    ContextSet,
    ContextShape,
    LogLevel,
    Json,
    Schema,
)

log = _internal_logging.InternalLogger(__name__)


__base_sdk: Optional[ReforgeSDK] = None
__options: Optional[Options] = None
__lock = _ReadWriteLock()


def set_options(options: Options) -> None:
    """Configure the SDK. SDK will be instantiated lazily with these options. Setting them again will have no effect unless reset_instance is called"""
    global __options
    with __lock.write_locked():
        __options = options


def get_sdk() -> ReforgeSDK:
    """Returns the singleton instance of the SDK. Created if needed using the options set by set_options"""
    global __base_sdk
    with __lock.read_locked():
        if __base_sdk:
            return __base_sdk

    with __lock.write_locked():
        if not __options:
            raise Exception("Options has not been set")
        if not __base_sdk:
            log.info(f"Initializing Reforge SDK version {version('reforge-python')}")
            __base_sdk = ReforgeSDK(__options)
            return __base_sdk


def reset_instance() -> None:
    """clears the singleton SDK instance so it will be recreated on the next get() call"""
    global __base_sdk
    global __lock
    __lock = _ReadWriteLock()
    old_sdk = __base_sdk
    __base_sdk = None
    if old_sdk:
        old_sdk.close()



import re
from collections import OrderedDict
from dataclasses import dataclass
import time

from ._internal_logging import (
    InternalLogger,
)
import requests
from requests import Response, RequestException
from requests.adapters import HTTPAdapter
from tenacity import (
    retry,
    stop_after_delay,
    wait_exponential,
    retry_if_exception_type,
)

import os

logger = InternalLogger(__name__)


def _get_version():
    try:
        version_file = os.path.join(os.path.dirname(__file__), "VERSION")
        with open(version_file, "r") as f:
            return f.read().strip()
    except (FileNotFoundError, IOError):
        return "development"


Version = _get_version()


VersionHeader = "X-Reforge-SDK-Version"

DEFAULT_TIMEOUT = 5  # seconds


# from https://findwork.dev/blog/advanced-usage-python-requests-timeouts-retries-hooks/
class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs) -> None:
        self.timeout = kwargs.pop("timeout", DEFAULT_TIMEOUT)
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs) -> Response:
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


class NoRetryAdapter(HTTPAdapter):
    def send(
        self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None
    ):
        return super().send(request, stream, timeout, verify, cert, proxies)


class UnauthorizedException(Exception):
    def __init__(self, api_key):
        api_key_prefix = api_key[:10] if api_key else ""
        super().__init__(
            f"Prefab attempts to fetch data are unauthorized using api key starting with {api_key_prefix}. Please check your api key."
        )


class HostIterator:
    def __init__(self, hosts):
        self.hosts = hosts
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if not self.hosts:
            raise StopIteration
        host = self.hosts[self.index]
        self.index = (self.index + 1) % len(self.hosts)
        return host


# --- Simple LRU Cache Implementation ---


@dataclass
class CacheEntry:
    data: bytes
    etag: str
    expires_at: float
    url: str  # The full URL from the successful response


class LRUCache:
    def __init__(self, max_size: int):
        self.max_size = max_size
        self.cache = OrderedDict()

    def get(self, key):
        try:
            value = self.cache.pop(key)
            self.cache[key] = value  # Mark as recently used.
            return value
        except KeyError:
            return None

    def set(self, key, value):
        if key in self.cache:
            self.cache.pop(key)
        elif len(self.cache) >= self.max_size:
            self.cache.popitem(last=False)
        self.cache[key] = value

    def clear(self):
        self.cache.clear()

    def __len__(self):
        return len(self.cache)


class ApiClient:
    def __init__(self, options):
        """
        :param options: An object with attributes such as:
            - prefab_api_urls: list of API host URLs (e.g. ["https://a.example.com", "https://b.example.com"])
            - version: version string
        """
        self.hosts = options.reforge_api_urls
        self.session = requests.Session()
        self.session.mount("https://", requests.adapters.HTTPAdapter())
        self.session.mount("http://", requests.adapters.HTTPAdapter())
        self.session.headers.update(
            {
                "X-Reforge-Client-Version": f"reforge-python-{getattr(options, 'version', 'development')}"
            }
        )
        # Initialize a cache (here with a maximum of 2 entries).
        self.cache = LRUCache(max_size=2)

    def get_host(self, attempt_number, host_list):
        return host_list[attempt_number % len(host_list)]

    def _get_attempt_number(self) -> int:
        """
        Retrieve the current attempt number from tenacity's statistics if available,
        otherwise default to 1.
        """
        stats = getattr(self.resilient_request, "statistics", None)
        if stats is None:
            return 1
        return stats.get("attempt_number", 1)

    def _build_url(self, path, hosts: list[str] = None) -> str:
        """
        Build the full URL using host-selection logic.
        """
        attempt_number = self._get_attempt_number()
        host = self.get_host(attempt_number - 1, hosts or self.hosts)
        return f"{host.rstrip('/')}/{path.lstrip('/')}"

    def _get_cached_response(self, url: str) -> Response:
        """
        If a valid cache entry exists for the given URL, return a synthetic Response.
        """
        now = time.time()
        entry = self.cache.get(url)
        if entry is not None and entry.expires_at > now:
            resp = Response()
            resp._content = entry.data
            resp.status_code = 200
            resp.headers = {"ETag": entry.etag, "X-Cache": "HIT"}
            resp.url = entry.url
            return resp
        return None

    def _apply_cache_headers(self, url: str, kwargs: dict) -> dict:
        """
        If a stale cache entry exists, add its ETag as an 'If-None-Match' header.
        """
        entry = self.cache.get(url)
        headers = kwargs.get("headers", {}).copy()
        if entry is not None and entry.etag:
            headers["If-None-Match"] = entry.etag
        kwargs["headers"] = headers
        return kwargs

    def _update_cache(self, url: str, response: Response) -> None:
        """
        If the response is cacheable (status 200, and Cache-Control does not include 'no-store'),
        update the cache. If Cache-Control includes 'no-cache', mark the cache entry as immediately expired,
        so that subsequent requests always trigger revalidation.
        """
        cache_control = response.headers.get("Cache-Control", "")
        if "no-store" in cache_control.lower():
            return

        etag = response.headers.get("ETag")
        max_age = 0
        m = re.search(r"max-age=(\d+)", cache_control)
        if m:
            max_age = int(m.group(1))

        # If 'no-cache' is present, then even though we may store the response,
        # we treat it as expired immediately so that every subsequent request is revalidated.
        if "no-cache" in cache_control.lower():
            expires_at = time.time()  # Immediately expired.
        else:
            expires_at = time.time() + max_age if max_age > 0 else 0

        if (etag is not None or max_age > 0) and expires_at > time.time():
            self.cache.set(
                url,
                CacheEntry(
                    data=response.content,
                    etag=etag,
                    expires_at=expires_at,
                    url=response.url,
                ),
            )
            response.headers["X-Cache"] = "MISS"

    def _send_request(self, method: str, url: str, **kwargs) -> Response:
        """
        Hook method to perform the actual HTTP request.
        """
        return self.session.request(method, url, **kwargs)

    @retry(
        stop=stop_after_delay(8),
        wait=wait_exponential(multiplier=1, min=0.05, max=2),
        retry=retry_if_exception_type((RequestException, ConnectionError, OSError)),
    )
    def resilient_request(
        self,
        path,
        method="GET",
        allow_cache: bool = False,
        hosts: list[str] = None,
        **kwargs,
    ) -> Response:
        """
        Makes a resilient (retrying) request.

        If allow_cache is True and the request method is GET, caching logic is applied.
        This includes:
          - Checking the cache and returning a synthetic response if valid.
          - Adding an 'If-None-Match' header when a stale entry exists.
          - Handling a 304 (Not Modified) response by returning the cached entry.
          - Caching a 200 response if Cache-Control permits.
        """
        url = self._build_url(path, hosts)
        if method.upper() == "GET" and allow_cache:
            cached = self._get_cached_response(url)
            if cached:
                return cached
            kwargs = self._apply_cache_headers(url, kwargs)
        response = self._send_request(method, url, **kwargs)
        if method.upper() == "GET" and allow_cache:
            if response.status_code == 304:
                cached = self.cache.get(url)
                if cached:
                    resp = Response()
                    resp._content = cached.data
                    resp.status_code = 200
                    resp.headers = {"ETag": cached.etag, "X-Cache": "HIT"}
                    resp.url = cached.url
                    return resp
            self._update_cache(url, response)
        return response



from __future__ import annotations
from typing import Optional

from .context import Context
from .constants import NoDefaultProvided, ConfigValueType
from ._internal_logging import InternalLogger

logger = InternalLogger(__name__)


class FeatureFlagSDK:
    def __init__(self, base_client):
        self.base_client = base_client

    def feature_is_on(
        self, feature_name, context: Optional[dict | Context] = None
    ) -> bool:
        return self.feature_is_on_for(feature_name, context)

    def feature_is_on_for(
        self, feature_name, context: Optional[dict | Context] = None
    ) -> bool:
        variant = self.base_client.config_sdk().get(
            feature_name, False, context=context
        )
        return self._is_on(variant)

    def get(
        self,
        feature_name,
        default=NoDefaultProvided,
        context: Optional[dict | Context] = None,
    ) -> ConfigValueType:
        return self._get(feature_name, default=default, context=context)

    def _get(
        self,
        feature_name,
        default=NoDefaultProvided,
        context: Optional[dict | Context] = None,
    ) -> ConfigValueType:
        return self.base_client.config_sdk().get(
            feature_name, default=default, context=context
        )

    def _is_on(self, variant) -> bool:
        try:
            if variant is None:
                return False
            if isinstance(variant, bool):
                return variant
            return variant.bool
        except Exception:
            logger.info(
                f"is_on methods only work for boolean feature flag variants. This feature flag's variant is '{variant}'. Returning False",
            )
            return False



import base64
import time
from typing import Optional, Callable

import sseclient  # type: ignore
from requests import Response

from sdk_reforge._internal_logging import InternalLogger
from sdk_reforge._requests import ApiClient, UnauthorizedException
import prefab_pb2 as Prefab
from sdk_reforge.config_sdk_interface import ConfigSDKInterface

SHORT_CONNECTION_THRESHOLD = 2  # seconds
CONSECUTIVE_SHORT_CONNECTION_LIMIT = 2  # times
MIN_BACKOFF_TIME = 1  # seconds
MAX_BACKOFF_TIME = 30  # seconds


class TooQuickConnectionException(Exception):
    pass


logger = InternalLogger(__name__)


class SSEConnectionManager:
    def __init__(
        self,
        api_client: ApiClient,
        config_client: ConfigSDKInterface,
        urls: list[str],
    ):
        self.api_client = api_client
        self.config_client = config_client
        self.sse_client: Optional[sseclient.SSEClient] = None
        self.timing = Timing()
        self.urls = urls

    def streaming_loop(self) -> None:
        too_short_connection_count = 0
        backoff_time = MIN_BACKOFF_TIME

        while self.config_client.continue_connection_processing():
            try:
                logger.debug("Starting streaming connection")
                headers = {
                    "Last-Event-ID": f"{self.config_client.highwater_mark()}",
                    "accept": "text/event-stream",
                }
                response = self.api_client.resilient_request(
                    "/api/v2/sse/config",
                    headers=headers,
                    stream=True,
                    auth=("authuser", self.config_client.options.api_key),
                    timeout=(5, 60),
                    hosts=self.urls,
                )
                response.raise_for_status()
                if response.ok:
                    elapsed_time = self.timing.time_execution(
                        lambda: self.process_response(response)
                    )
                    if elapsed_time < SHORT_CONNECTION_THRESHOLD:
                        too_short_connection_count += 1
                        if (
                            too_short_connection_count
                            >= CONSECUTIVE_SHORT_CONNECTION_LIMIT
                        ):
                            raise TooQuickConnectionException()
                    else:
                        too_short_connection_count = 0
                        backoff_time = MIN_BACKOFF_TIME
                    time.sleep(backoff_time)
            except UnauthorizedException:
                self.config_client.handle_unauthorized_response()
            except TooQuickConnectionException as e:
                logger.debug(f"Connection ended quickly: {str(e)}. Will apply backoff.")
                backoff_time = min(backoff_time * 2, MAX_BACKOFF_TIME)
                time.sleep(backoff_time)
            except Exception as e:
                if not self.config_client.is_shutting_down():
                    logger.warning(
                        f"Streaming connection error: {str(e)}, Will retry in {backoff_time} seconds"
                    )
                    backoff_time = min(backoff_time * 2, MAX_BACKOFF_TIME)
                    time.sleep(backoff_time)

    """
    Hand off a successful response here for processing
    """

    def process_response(self, response: Response) -> None:
        self.sse_client = sseclient.SSEClient(response)
        if self.sse_client is not None:
            for event in self.sse_client.events():
                if self.config_client.is_shutting_down():
                    logger.info("Client is shutting down, exiting SSE event loop")
                    return
                if event.data:
                    configs = Prefab.Configs.FromString(base64.b64decode(event.data))
                    self.config_client.load_configs(configs, "sse_streaming")
            self.sse_client.close()
            self.sse_client = None


class Timing:
    def time_execution(self, func: Callable[[], None]) -> float:
        """Executes the given function and returns the time it took to execute."""
        start_time = self.now()
        func()  # Execute the block of code
        return self.now() - start_time

    def now(self) -> float:
        """Get the current time. This can be mocked in tests."""
        return time.time()



import threading
import logging
from typing import Iterator

import prefab_pb2 as Prefab


LLV = Prefab.LogLevel.Value


python_log_level_name_to_prefab_log_levels = {
    "debug": LLV("DEBUG"),
    "info": LLV("INFO"),
    "warn": LLV("WARN"),
    "warning": LLV("WARN"),
    "error": LLV("ERROR"),
    "critical": LLV("FATAL"),
}

python_to_prefab_log_levels = {
    logging.NOTSET: LLV("DEBUG"),
    logging.DEBUG: LLV("DEBUG"),
    logging.INFO: LLV("INFO"),
    logging.WARN: LLV("WARN"),
    logging.ERROR: LLV("ERROR"),
    logging.CRITICAL: LLV("FATAL"),
}

prefab_to_python_log_levels = {
    LLV("TRACE"): logging.DEBUG,
    LLV("DEBUG"): logging.DEBUG,
    LLV("INFO"): logging.INFO,
    LLV("WARN"): logging.WARN,
    LLV("ERROR"): logging.ERROR,
    LLV("FATAL"): logging.CRITICAL,
}


def iterate_dotted_string(s: str) -> Iterator[str]:
    parts = s.split(".")
    for i in range(len(parts), 0, -1):
        yield ".".join(parts[:i])


class ReentrancyCheck:
    thread_local = threading.local()

    @staticmethod
    def set() -> None:
        ReentrancyCheck.thread_local.prefab_log_reentrant = True

    @staticmethod
    def is_set() -> bool:
        # Safely check if the thread-local variable is set and return True/False
        return getattr(ReentrancyCheck.thread_local, "prefab_log_reentrant", False)

    @staticmethod
    def clear() -> None:
        try:
            # Attempt to delete the variable
            delattr(ReentrancyCheck.thread_local, "prefab_log_reentrant")
        except AttributeError:
            # Variable was not set for this thread
            pass


class InternalLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)
        self.thread_local = threading.local()

    def log(self, level: int, msg, *args, **kwargs) -> None:
        if not ReentrancyCheck.is_set():
            extras = kwargs.pop("extra", {})
            extras["prefab_internal"] = True
            # Pass the possibly-modified 'extra' dictionary to the underlying logger
            super().log(level, msg, *args, extra=extras, **kwargs)



from __future__ import annotations
import functools
import threading
from urllib3 import Retry


from ._telemetry import TelemetryManager
from ._internal_logging import InternalLogger
from .context import Context, ScopedContext
from .config_sdk import ConfigSDK
from .feature_flag_sdk import FeatureFlagSDK
from .options import Options
from ._requests import TimeoutHTTPAdapter, VersionHeader, Version
from typing import Optional
import prefab_pb2 as Prefab
import uuid
import requests
from urllib.parse import urljoin
from .constants import (
    NoDefaultProvided,
    ConfigValueType,
    ContextDictOrContext,
    PostBodyType,
)

logger = InternalLogger(__name__)


class ReforgeSDK:
    max_sleep_sec = 10
    base_sleep_sec = 0.5

    def __init__(self, options: Options) -> None:
        self.shutdown_flag = threading.Event()
        self.options = options
        self.global_context = options.global_context
        self.instance_hash = str(uuid.uuid4())
        self.telemetry_manager = TelemetryManager(self, options)
        if not options.is_local_only():
            self.telemetry_manager.start_periodic_sync()
        self.api_urls = options.reforge_api_urls
        # Define the retry strategy
        retry_strategy = Retry(
            total=2,  # Maximum number of retries
            status_forcelist=[429, 500, 502, 503, 504],  # HTTP status codes to retry on
            allowed_methods=["POST", "GET"],
        )
        # Create an TimeoutHTTPAdapter adapter with the retry strategy and a standard timeout and mount it to session
        adapter = TimeoutHTTPAdapter(max_retries=retry_strategy, timeout=5)
        self.session = requests.Session()
        self.session.mount("https://", adapter)
        self.session.headers.update({VersionHeader: f"sdk-python-{Version}"})
        if options.is_local_only():
            logger.info(f"Reforge SDK {Version} running in local-only mode")
        else:
            logger.info(
                f"Reforge SDK {Version} connecting to %s, secure %s"
                % (
                    options.reforge_api_urls,
                    options.http_secure,
                ),
            )

        self.context().clear()
        self.config_sdk()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get(
        self,
        key: str,
        default: ConfigValueType = NoDefaultProvided,
        context: Optional[ContextDictOrContext] = None,
    ) -> ConfigValueType:
        if self.is_ff(key):
            return self.feature_flag_sdk().get(key, default=default, context=context)
        else:
            return self.config_sdk().get(key, default=default, context=context)

    def enabled(
        self, feature_name: str, context: Optional[ContextDictOrContext] = None
    ) -> bool:
        return self.feature_flag_sdk().feature_is_on_for(feature_name, context=context)

    def is_ff(self, key: str) -> bool:
        raw = self.config_sdk().config_resolver.raw(key)
        if raw is not None and raw.config_type == Prefab.ConfigType.Value(
            "FEATURE_FLAG"
        ):
            return True
        return False

    def context(self) -> Context:
        return Context.get_current()

    @staticmethod
    def scoped_context(context: dict | Context) -> ScopedContext:
        return Context.scope(context)

    @functools.cache
    def config_sdk(self) -> ConfigSDK:
        client = ConfigSDK(self)
        return client

    @functools.cache
    def feature_flag_sdk(self) -> FeatureFlagSDK:
        return FeatureFlagSDK(self)

    def post(self, path: str, body: PostBodyType) -> requests.models.Response:
        headers = {
            "Content-Type": "application/x-protobuf",
            "Accept": "application/x-protobuf",
        }

        endpoint = urljoin(self.options.telemetry_url or "", path)

        return self.session.post(
            endpoint,
            headers=headers,
            data=body.SerializeToString(),
            auth=("authuser", self.options.api_key or ""),
        )

    def is_ready(self) -> bool:
        return self.config_sdk().is_ready()

    def set_global_context(
        self, global_context: Optional[ContextDictOrContext] = None
    ) -> "ReforgeSDK":
        self.global_context = Context.normalize_context_arg(global_context)
        return self

    def close(self) -> None:
        if not self.shutdown_flag.is_set():
            logger.info("Shutting down prefab client instance")
            self.shutdown_flag.set()
            self.config_sdk().close()
        else:
            logger.warning("Close already called")
